import asyncio
import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from importlib.metadata import version

from acp_sdk.client import Client
from acp_sdk.models import (
    AgentName,
    AwaitResume,
    Message,
    Run,
    RunId,
    RunStatus,
    SessionId,
)
from mcp.server import Server
from mcp.server.lowlevel.helper_types import ReadResourceContents
from mcp.types import (
    EmbeddedResource,
    ImageContent,
    Resource,
    TextContent,
    TextResourceContents,
    Tool,
)
from pydantic import AnyHttpUrl, AnyUrl, BaseModel, TypeAdapter

logger = logging.getLogger(__name__)


class RunAgentInput(BaseModel):
    agent: AgentName
    input: list[Message]
    session: SessionId | None = None


class RunAgentResumeInput(BaseModel):
    run: RunId
    await_resume: AwaitResume


class EmptySchema(BaseModel):
    pass


@dataclass
class Context:
    client: Client


def create_adapter(acp_url: AnyHttpUrl) -> Server:
    @asynccontextmanager
    async def lifespan(server: Server) -> AsyncIterator[Context]:
        logger.info("Running server")
        try:
            async with Client(base_url=str(acp_url)) as client:
                yield Context(client=client)
        finally:
            logger.info("Server shutdown")

    server = Server("acp-mcp", version=version("acp-mcp"), lifespan=lifespan)

    @server.list_resources()
    async def list_resources() -> list[Resource]:
        client = server.request_context.lifespan_context.client
        agents = [agent async for agent in client.agents()]
        return [
            Resource(
                uri=_create_agent_uri(acp_url, agent.name),
                name=agent.name,
                description=agent.description,
            )
            for agent in agents
        ]

    @server.read_resource()
    async def read_resource(uri: AnyUrl) -> list[ReadResourceContents]:
        agent = _parse_agent_from_url(uri)
        if agent is None:
            raise ValueError("Invalid resource")

        client = server.request_context.lifespan_context.client
        agent = await client.agent(name=agent)
        return [
            ReadResourceContents(
                mime_type="application/json",
                content=agent.model_dump_json(),
            )
        ]

    @server.list_tools()
    async def list_tools() -> list[Tool]:
        return [
            Tool(
                name="list_agents",
                description="Lists available agents",
                inputSchema=EmptySchema.model_json_schema(),
            ),
            Tool(
                name="run_agent",
                description="Runs an agent",
                inputSchema=RunAgentInput.model_json_schema(),
            ),
            Tool(
                name="resume_run_agent",
                description="Resumes an agent run",
                inputSchema=RunAgentResumeInput.model_json_schema(),
            ),
        ]

    @server.call_tool()
    async def handle_call_tool(
        name: str, arguments: dict | None
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        client = server.request_context.lifespan_context.client
        match name:
            case "list_agents":
                return [
                    EmbeddedResource(
                        type="resource",
                        resource=TextResourceContents(
                            uri=_create_agent_uri(acp_url, agent.name),
                            mimeType="application/json",
                            text=agent.model_dump_json(),
                        ),
                    )
                    async for agent in client.agents()
                ]
            case "run_agent":
                input = RunAgentInput.model_validate(arguments)
                async with client.session(session_id=input.session) as session:
                    run = await session.run_sync(input.input, agent=input.agent)
                    return [TextContent(type="text", text=_run_to_tool_text(run))]
            case "resume_run_agent":
                input = RunAgentResumeInput.model_validate(arguments)
                run = await client.run_resume_sync(
                    input.await_resume,
                    run_id=input.run,
                )
                return [TextContent(type="text", text=_run_to_tool_text(run))]
            case _:
                raise ValueError("Invalid tool name")

    return server


def _create_agent_uri(base_url: AnyHttpUrl, agent: AgentName) -> AnyHttpUrl:
    return AnyHttpUrl.build(
        scheme=base_url.scheme,
        username=base_url.username,
        password=base_url.password,
        host=base_url.host,
        port=base_url.port,
        path=(base_url.path.rstrip("/") + f"/agents/{agent}").lstrip("/"),
        query=base_url.query,
        fragment=base_url.fragment,
    )


def _parse_agent_from_url(url: AnyUrl) -> AgentName | None:
    path_segments = url.path.split("/")
    if len(path_segments) < 3 or path_segments[1] != "agents" or len(path_segments[2]) == 0:
        return None
    return path_segments[2]


def _run_to_tool_text(run: Run) -> str:
    "Encodes run into tool response"
    match run.status:
        case RunStatus.AWAITING:
            return f"Run {run.run_id} awaits: {run.await_request.model_dump_json()}"
        case RunStatus.COMPLETED:
            return TypeAdapter(list[Message]).dump_json(run.output).decode()
        case RunStatus.CANCELLED:
            raise asyncio.CancelledError("Agent run cancelled")
        case RunStatus.FAILED:
            raise RuntimeError("Agent failed with error:", run.error)
        case _:
            raise RuntimeError(f"Agent {run.status.value}")



import asyncio
import logging

from mcp import stdio_server
from mcp.server.lowlevel import NotificationOptions
from mcp.server.models import InitializationOptions
from pydantic import AnyHttpUrl

from acp_mcp.adapter import create_adapter

logger = logging.getLogger(__name__)


async def serve(acp_url: AnyHttpUrl) -> None:
    server = create_adapter(acp_url=acp_url)
    async with stdio_server() as (read_stream, write_stream):
        await server.run(
            read_stream,
            write_stream,
            InitializationOptions(
                server_name=server.name,
                server_version=server.version,
                capabilities=server.get_capabilities(
                    notification_options=NotificationOptions(), experimental_capabilities={}
                ),
            ),
        )


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser(prog="acp-mcp", description="Serve ACP agents over MCP")
    parser.add_argument("url", type=AnyHttpUrl, help="The URL of an ACP server")
    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level",
    )

    args = parser.parse_args()

    logging.basicConfig(level=args.log_level, format="%(asctime)s - %(levelname)s - %(message)s")

    asyncio.run(serve(acp_url=args.url))


if __name__ == "__main__":
    main()



import pytest
from mcp import ClientSession
from mcp.types import TextContent

from e2e.config import Config

mock_agents = [
    f"http://localhost:{Config.PORT}/agents/echo",
    f"http://localhost:{Config.PORT}/agents/slow_echo",
]


@pytest.mark.anyio
async def test_list_tools(session: ClientSession) -> None:
    expected_tool_names = ["list_agents", "run_agent", "resume_run_agent"]

    result = await session.list_tools()
    tool_names = [tool.name for tool in result.tools]

    for name in expected_tool_names:
        assert name in tool_names


@pytest.mark.anyio
async def test_list_agents(session: ClientSession) -> None:
    result = await session.call_tool("list_agents")
    assert not result.isError
    assert len(result.content) == len(mock_agents)


@pytest.mark.anyio
async def test_run_agent(session: ClientSession) -> None:
    result = await session.call_tool("run_agent", {"agent": "echo", "input": [{"parts": [{"content": "Howdy!"}]}]})
    assert not result.isError
    assert isinstance(result.content[0], TextContent)



import pytest
from mcp import ClientSession

from e2e.config import Config

mock_agents = [
    f"http://localhost:{Config.PORT}/agents/echo",
    f"http://localhost:{Config.PORT}/agents/slow_echo",
]


@pytest.mark.anyio
async def test_list_agents(session: ClientSession) -> None:
    result = await session.list_resources()
    resource_uris = [str(resource.uri) for resource in result.resources]

    for agent in mock_agents:
        assert agent in resource_uris


@pytest.mark.anyio
@pytest.mark.parametrize(
    "agent",
    mock_agents,
)
async def test_read_agent(session: ClientSession, agent: str) -> None:
    result = await session.read_resource(agent)
    assert str(result.contents[0].uri) == agent



import pytest
from acp_sdk.models import AgentName
from pydantic import AnyHttpUrl

from acp_mcp.adapter import _create_agent_uri, _parse_agent_from_url


@pytest.mark.parametrize(
    "base,agent,expected",
    [
        (AnyHttpUrl("http://localhost:8000"), "foobar", AnyHttpUrl("http://localhost:8000/agents/foobar")),
        (AnyHttpUrl("http://localhost:8000/"), "foobar", AnyHttpUrl("http://localhost:8000/agents/foobar")),
        (
            AnyHttpUrl("https://localhost:8000/v1/acp"),
            "foobar",
            AnyHttpUrl("https://localhost:8000/v1/acp/agents/foobar"),
        ),
    ],
)
def test_create_agent_uri(base: AnyHttpUrl, agent: AgentName, expected: AnyHttpUrl) -> None:
    assert _create_agent_uri(base, agent) == expected


@pytest.mark.parametrize(
    "agent_uri,expected",
    [
        (AnyHttpUrl("http://localhost:8000/agents/foobar"), "foobar"),
        (AnyHttpUrl("http://localhost:8000/agents/"), None),
        (AnyHttpUrl("http://localhost:8000"), None),
        (AnyHttpUrl("http://localhost:8000/tools/foobat"), None),
    ],
)
def test_parse_agent_from_url(agent_uri: AnyHttpUrl, expected: AgentName) -> None:
    assert _parse_agent_from_url(agent_uri) == expected



import asyncio
from collections.abc import AsyncIterator
from threading import Thread

import pytest
from acp_sdk.models import (
    Message,
)
from acp_sdk.server import Context
from acp_sdk.server import Server as ACPServer
from mcp import ClientSession
from mcp.shared.memory import create_connected_server_and_client_session
from pydantic import AnyHttpUrl

from acp_mcp.adapter import create_adapter
from e2e.config import Config


@pytest.fixture
async def session() -> AsyncIterator[ClientSession]:
    acp_server = ACPServer()

    @acp_server.agent()
    async def echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        "Echoes everything"
        for message in input:
            yield message

    @acp_server.agent()
    async def slow_echo(input: list[Message], context: Context) -> AsyncIterator[Message]:
        "Echoes everything but slowly"
        for message in input:
            await asyncio.sleep(1)
            yield message

    thread = Thread(target=acp_server.run, kwargs={"port": Config.PORT}, daemon=True)
    thread.start()

    await asyncio.sleep(1)

    server = create_adapter(AnyHttpUrl(f"http://localhost:{Config.PORT}"))
    async with create_connected_server_and_client_session(server=server) as session:
        yield session

    acp_server.should_exit = True
    thread.join(timeout=2)
